# Generated by Django 2.1.4 on 2018-12-09 19:11

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import magicauth.utils


class Migration(migrations.Migration):

    initial = True

    dependencies = [migrations.swappable_dependency(settings.AUTH_USER_MODEL)]

    operations = [
        migrations.CreateModel(
            name="MagicToken",
            fields=[
                (
                    "key",
                    models.CharField(
                        default=magicauth.utils.generate_token,
                        max_length=255,
                        primary_key=True,
                        serialize=False,
                        verbose_name="Key",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="magic_token",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "Magic Token",
                "verbose_name_plural": "Magic Tokens",
                "ordering": ("-created",),
            },
        )
    ]



import math

from django.contrib.auth import get_user_model
from django.contrib.sites.shortcuts import get_current_site
from django.core.mail import send_mail
from django.template import loader

from magicauth import settings as magicauth_settings
from magicauth.models import MagicToken


class SendTokenMixin(object):
    """
    Helper for sending an email containing a link containing the MagicToken.
    """

    def create_token(self, user):
        token = MagicToken.objects.create(user=user)
        return token

    def get_user_from_email(self, user_email):
        """
        Query the DB for the user corresponding to the email.
         - We use get_user_model() instead of User (in case the Django app has customised the User
        class)
         - We use magicauth_settings.EMAIL_FIELD, which is the name of the field in the user
        model. By default "username" but not always.
        """
        user_class = get_user_model()
        email_field = magicauth_settings.EMAIL_FIELD
        field_lookup = {f"{email_field}__iexact": user_email}
        user = user_class.objects.get(**field_lookup)
        return user

    def send_email(self, user, user_email, token, extra_context=None):
        email_subject = magicauth_settings.EMAIL_SUBJECT
        html_template = magicauth_settings.EMAIL_HTML_TEMPLATE
        text_template = magicauth_settings.EMAIL_TEXT_TEMPLATE
        from_email = magicauth_settings.FROM_EMAIL
        context = {
            "token": token,
            "user": user,
            "site": get_current_site(self.request),
            "TOKEN_DURATION_MINUTES": math.floor(magicauth_settings.TOKEN_DURATION_SECONDS / 60),
            "TOKEN_DURATION_SECONDS": magicauth_settings.TOKEN_DURATION_SECONDS,
        }
        if extra_context:
            context.update(extra_context)
        text_message = loader.render_to_string(text_template, context)
        html_message = loader.render_to_string(html_template, context)
        send_mail(
            subject=email_subject,
            message=text_message,
            from_email=from_email,
            html_message=html_message,
            recipient_list=[user_email],
            fail_silently=False,
        )

    def send_token(self, user_email, extra_context=None):
        user = self.get_user_from_email(user_email)
        token = self.create_token(user)
        self.send_email(user, user_email, token, extra_context)



from django import forms
from django.contrib.auth import get_user_model
from django.utils.module_loading import import_string
from django.core.validators import RegexValidator
from django.core.exceptions import ValidationError

from django_otp import user_has_device, devices_for_user

from magicauth import settings as magicauth_settings
from magicauth.models import MagicToken

email_unknown_callback = import_string(magicauth_settings.EMAIL_UNKNOWN_CALLBACK)


class EmailForm(forms.Form):
    email = forms.EmailField()

    def clean_email(self):
        user_email = self.cleaned_data["email"]
        user_email = user_email.lower()

        email_field = magicauth_settings.EMAIL_FIELD
        field_lookup = {f"{email_field}__iexact": user_email}
        if not get_user_model().objects.filter(**field_lookup).exists():
            email_unknown_callback(user_email)
        return user_email


class OTPForm(forms.Form):
    OTP_NUM_DIGITS = magicauth_settings.OTP_NUM_DIGITS
    otp_token = forms.CharField(
        max_length=OTP_NUM_DIGITS,
        min_length=OTP_NUM_DIGITS,
        validators=[RegexValidator(r"^\d{6}$")],
        label=f"Entrez le code à {OTP_NUM_DIGITS} chiffres généré par votre téléphone ou votre carte OTP",
        widget=forms.TextInput(attrs={"autocomplete": "off"}),
    )

    def __init__(self, user, *args, **kwargs):
        super(OTPForm, self).__init__(*args, **kwargs)
        self.user = user

    def clean_otp_token(self):
        otp_token = self.cleaned_data["otp_token"]
        user = self.user
        if not user_has_device(user):
            raise ValidationError("Le système n'a pas trouvé d'appareil (carte OTP ou générateur sur téléphone) pour votre compte. Contactez le support pour en ajouter un.")

        for device in devices_for_user(user):
            if device.verify_is_allowed() and device.verify_token(otp_token):
                return otp_token

        raise ValidationError("Ce code n'est pas valide.")




from pytest import mark

from django.core import mail
from django.shortcuts import reverse
from magicauth import settings
from magicauth.models import MagicToken
from tests import factories

'''
Step 2 of login process : see doc in magicauth/views.py for step details
'''

pytestmark = mark.django_db


def post_email(client, email):
    url = reverse("magicauth-login")
    return client.post(url, data={"email": email})


def test_posting_email_for_valid_existing_user_redirects(client):
    user = factories.UserFactory()
    response = post_email(client, user.email)
    assert response.status_code == 302


def test_posting_email_for_valid_existing_user_sends_email(client):
    user = factories.UserFactory()
    post_email(client, user.email)
    assert len(mail.outbox) == 1


def test_posting_email_sends_email_with_redirection_to_default(client):
    user = factories.UserFactory()
    post_email(client, user.email)
    assert "?next=/landing/" in mail.outbox[0].body


def test_posting_email_sends_email_with_redirection_to_next(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login") + "?next=/test_dashboard/"
    client.post(url, data={"email": user.email})
    assert "?next=/test_dashboard/" in mail.outbox[0].body


def test_posting_email_with_usafe_next_does_not_send_email_and_returns_404(client):
    user = factories.UserFactory()
    url = reverse("magicauth-login") + "?next=http://www.myfishingsite.com/"
    response = client.post(url, data={"email": user.email})
    assert len(mail.outbox) == 0
    assert response.status_code == 404


def test_posting_email_for_valid_existing_user_creates_token(client):
    user = factories.UserFactory()
    count_before = MagicToken.objects.count()
    post_email(client, user.email)
    count_after = MagicToken.objects.count()
    assert count_after == count_before + 1


def test_loging_with_email_is_case_insensitive(client):
    user = factories.UserFactory()
    response = post_email(client, user.email.upper())
    assert response.status_code == 302
    assert len(mail.outbox) == 1


def test_posting_unknown_email_raise_error(client):
    response = post_email(client, "unknown@email.com")
    assert "invalid" in str(response.content)


def test_posting_unknown_email_does_not_send_email(client):
    post_email(client, "unknown@email.com")
    assert len(mail.outbox) == 0

# Tests with OTPs
def post_email_and_OTP(client, email, OTP):
    url = reverse("magicauth-login")
    data = {"email": email, "otp_token": OTP}
    return client.post(url, data=data)

def test_posting_good_email_and_good_totp_success(client):
    settings.ENABLE_2FA = True
    token = factories.MagicTokenFactory()
    thierry = token.user
    device = thierry.staticdevice_set.create()
    device.token_set.create(token="123456")

    response = post_email_and_OTP(client, thierry.email, "123456")

    assert response.status_code == 302
    assert len(mail.outbox) == 1


def test_posting_good_email_and_wrong_otp_error(client):
    settings.ENABLE_2FA = True
    token = factories.MagicTokenFactory()
    thierry = token.user
    device = thierry.staticdevice_set.create()
    device.token_set.create(token="123456")

    response = post_email_and_OTP(client, thierry.email, "567654")

    assert response.status_code == 200
    assert "Ce code n&#x27;est pas valide." in str(response.content)
    assert len(mail.outbox) == 0


def test_posting_wrong_email_and_wrong_otp_error(client):
    settings.ENABLE_2FA = True
    token = factories.MagicTokenFactory()
    thierry = token.user
    device = thierry.staticdevice_set.create()
    device.token_set.create(token="123456")

    response = post_email_and_OTP(client, "unknown@email.com", "567654")

    assert response.status_code == 200
    assert "invalid" in str(response.content)
    assert len(mail.outbox) == 0

def test_thierry_has_several_devices_first_device(client):
    settings.ENABLE_2FA = True
    token = factories.MagicTokenFactory()
    thierry = token.user
    device_1 = thierry.staticdevice_set.create()
    device_1.token_set.create(token="123456")
    device_2 = thierry.staticdevice_set.create()
    device_2.token_set.create(token="111111")

    response = post_email_and_OTP(client, thierry.email, "123456")

    assert response.status_code == 302
    assert len(mail.outbox) == 1

def test_thierry_has_several_devices_second_device(client):
    settings.ENABLE_2FA = True
    token = factories.MagicTokenFactory()
    thierry = token.user
    device_1 = thierry.staticdevice_set.create()
    device_1.token_set.create(token="123456")
    device_2 = thierry.staticdevice_set.create()
    device_2.token_set.create(token="111111")

    response = post_email_and_OTP(client, thierry.email, "111111")

    assert response.status_code == 302
    assert len(mail.outbox) == 1



from pytest import mark

from django.shortcuts import reverse
from tests import factories

'''
Step 3 of login process : see doc in magicauth/views.py for step details
'''

pytestmark = mark.django_db


def test_email_sent_page_loads(client):
    url = reverse("magicauth-email-sent")
    response = client.get(url)
    assert response.status_code == 200


def test_email_sent_page_raises_404_if_unsafe_next_url(client):
    url = reverse("magicauth-email-sent") + "?next=http://www.myfishingsite.com/"
    response = client.get(url)
    assert response.status_code == 404


def test_email_sent_page_raises_404_if_unsafe_next_url_with_authenticated_user(client):
    user = factories.UserFactory()
    client.force_login(user)
    url = reverse("magicauth-email-sent") + "?next=http://www.myfishingsite.com/"
    response = client.get(url)
    assert response.status_code == 404
    assert user.is_authenticated



SECRET_KEY = "can you keep a secret?"

DATABASES = {"default": {"ENGINE": "django.db.backends.sqlite3"}}

ROOT_URLCONF = "tests.test_url"

INSTALLED_APPS = [
    "tests",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "magicauth",
    "django_otp",
    "django_otp.plugins.otp_static",
]

SESSION_ENGINE = "django.contrib.sessions.backends.cache"

MAGICAUTH_FROM_EMAIL = "user@domain.user"
MAGICAUTH_LOGGED_IN_REDIRECT_URL_NAME = "test_home"


TEMPLATES = [
    {"BACKEND": "django.template.backends.django.DjangoTemplates", "APP_DIRS": True}
]

MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.sites.middleware.CurrentSiteMiddleware",
    "django_otp.middleware.OTPMiddleware",
]



import urllib.parse
import logging

from django.http import Http404
from django.urls import reverse
try:
    from django.utils.http import url_has_allowed_host_and_scheme
except ImportError:
    # For Django < v3, use old deprecated function
    from django.utils.http import is_safe_url as url_has_allowed_host_and_scheme

from magicauth import settings as magicauth_settings


logger = logging.getLogger()


class NextUrlMixin(object):
    """
    Helper for managing the 'next url' parameter.
    """

    def get_next_url(self, request):
        """
        Get the next url from the querystring parameters (?next=/my/next/page).
        If the next parameter is not there, returns the default redirect url
        """
        next_url = request.GET.get("next")
        if not next_url:
            next_url = reverse(magicauth_settings.LOGGED_IN_REDIRECT_URL_NAME)
        if not url_has_allowed_host_and_scheme(next_url, allowed_hosts={request.get_host()}, require_https=True):
            # We are not logging the unsafe URL to prevent code injections in logs
            logger.warning("[MagicAuth] an unsafe URL was used through a login link")
            raise Http404
        return next_url

    def get_next_url_encoded(self, request):
        """
        Use this when the URL needs to be encoded, for instance when including the URL to string
        before a redirect.
        """
        url = self.get_next_url(self.request)
        return urllib.parse.quote(url)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        next_url = self.get_next_url(self.request)
        context["next_url"] = next_url
        return context



from datetime import timedelta
from pytest import mark
import urllib.parse

from django.shortcuts import reverse
from django.utils import timezone
from magicauth.models import MagicToken
from magicauth import settings
from tests import factories

'''
Step 5 of login process : see doc in magicauth/views.py for step details
'''

pytestmark = mark.django_db


def open_magic_link(client, token, next=None):
    url = reverse("magicauth-validate-token", args=[token.key])
    if next:
        # Encode the url (with urllib.parse.quote) otherwise URL params get lost.
        url += '?next=' + urllib.parse.quote(next)
    return client.get(url)


def test_opening_magic_link_triggers_login(client):
    token = factories.MagicTokenFactory()
    open_magic_link(client, token)
    assert "_auth_user_id" in client.session


def test_opening_magic_link_with_valid_token_redirects(client):
    token = factories.MagicTokenFactory()
    response = open_magic_link(client, token)
    assert response.status_code == 302
    assert response.url == "/landing/"


def test_opening_magic_link_with_a_next_sets_a_new_url(client):
    token = factories.MagicTokenFactory()
    next_url_raw = "/test_dashboard/?a=test&b=test"
    response = open_magic_link(client, token, next_url_raw)
    assert response.status_code == 302
    assert response.url == next_url_raw


def test_validate_token_view_with_unsafe_next_does_not_log_in(client):
    token = factories.MagicTokenFactory()
    next_url = 'http://www.myfishingsite.com/'
    open_magic_link(client, token, next_url)
    assert "_auth_user_id" not in client.session


def test_validate_token_view_with_unsafe_next_raises_404(client):
    token = factories.MagicTokenFactory()
    next_url = 'http://www.myfishingsite.com/'
    response = open_magic_link(client, token, next_url)
    assert response.status_code == 404


def test_validate_token_view_with_unsafe_next_raises_404_for_loggedin_user(client):
    token = factories.MagicTokenFactory()
    user = factories.UserFactory()
    client.force_login(user)
    next_url = 'http://www.myfishingsite.com/'
    response = open_magic_link(client, token, next_url)
    assert response.status_code == 404
    assert user.is_authenticated


def test_token_is_removed_after_visiting_magic_link(client):
    token = factories.MagicTokenFactory()
    count_before = MagicToken.objects.count()
    open_magic_link(client, token)
    count_after = MagicToken.objects.count()
    assert count_after == count_before - 1


def test_duplicate_token_for_same_user_is_removed_after_visiting_magic_link(client):
    token = factories.MagicTokenFactory()
    duplicate = factories.MagicTokenFactory(user=token.user)
    open_magic_link(client, token)
    assert duplicate not in MagicToken.objects.all()


def test_unknown_token_does_not_login(client):
    url = reverse("magicauth-validate-token", args=["unknown-token"])
    client.get(url)
    assert "_auth_user_id" not in client.session


def test_unknown_token_redirects_to_login(client):
    url = reverse("magicauth-validate-token", args=["unknown-token"])
    response = client.get(url)
    assert response.status_code == 302
    assert response.url == '/login/'


def create_expired_token():
    token = factories.MagicTokenFactory()
    token.created = timezone.now() - timedelta(seconds=(settings.TOKEN_DURATION_SECONDS * 2))
    token.save()
    return token


def test_expired_token_does_not_login(client):
    token = create_expired_token()
    open_magic_link(client, token)
    assert "_auth_user_id" not in client.session


def test_expired_token_redirects_to_login(client):
    token = create_expired_token()
    response = open_magic_link(client, token)
    assert response.status_code == 302
    assert response.url == '/login/'


def test_expired_token_is_deleted_when_visited(client):
    token = create_expired_token()
    open_magic_link(client, token)
    assert token not in MagicToken.objects.all()


def test_expired_token_is_deleted_when_valid_token_is_visited(client):
    expired_token = create_expired_token()
    valid_token = factories.MagicTokenFactory(user=expired_token.user)
    open_magic_link(client, valid_token)
    assert expired_token not in MagicToken.objects.all()



from pytest import mark

from django.shortcuts import reverse
from magicauth import settings
import urllib.parse
from tests import factories


'''
Step 4 of login process : see doc in magicauth/views.py for step details

Note : We do not test that the javascript actually does the redirect. We just test the django template's context.
'''

pytestmark = mark.django_db

def open_magic_link_with_wait(client, token, next=None):
    url = reverse("magicauth-wait", kwargs={"key": token.key})
    if next:
        # Encode the url (with urllib.parse.quote) otherwise URL params get lost.
        url += '?next=' + urllib.parse.quote(next)
    return client.get(url)


def test_wait_page_loads(client):
    url = reverse("magicauth-wait", kwargs={"key": 'some-token'})
    response = client.get(url)
    assert response.status_code == 200


def test_wait_page_will_redirect_to_validate_token(client):
    token = factories.MagicTokenFactory()
    response = open_magic_link_with_wait(client, token)

    redirect_url = reverse('magicauth-validate-token', kwargs={"key": token.key})
    assert redirect_url in response.context_data['next_step_url']


def test_wait_page_will_redirect_with_next_param(client):
    token = factories.MagicTokenFactory()
    response = open_magic_link_with_wait(client, token, '/test_dashboard/')

    assert 'next=/test_dashboard/' in response.context_data['next_step_url']


def test_wait_page_will_redirect_with_default_next_param(client):
    token = factories.MagicTokenFactory()
    response = open_magic_link_with_wait(client, token)

    assert 'next=/landing/' in response.context_data['next_step_url']


def test_wait_page_will_redirect_in_WAIT_SECONDS(client):
    token = factories.MagicTokenFactory()
    response = open_magic_link_with_wait(client, token)

    assert response.context_data['WAIT_SECONDS'] == settings.WAIT_SECONDS


def test_wait_page_raises_404_if_unsafe_next_url(client):
    token = factories.MagicTokenFactory()
    response = open_magic_link_with_wait(
        client, token, 'http://www.myfishingsite.com/')

    assert response.status_code == 404



import binascii
import os


from django import forms

from . import settings as magicauth_settings


def generate_token():
    return binascii.hexlify(os.urandom(20)).decode()


def raise_error(email=None):
    """
    Just raise an error - this can be used as a call back function
    when no user was found in DB during the login process.
    """
    raise forms.ValidationError(magicauth_settings.EMAIL_UNKNOWN_MESSAGE)



from django.contrib import admin

from .models import MagicToken


@admin.register(MagicToken)
class MagicTokenAdmin(admin.ModelAdmin):
    list_display = ("key", "user", "created")
    raw_id_fields = ("user",)



from django.conf import settings as django_settings

# To add magicauth to your site, you need to add these values to your site settings
# (the rest have defaults, you are free to change them if you want):
# MAGICAUTH_FROM_EMAIL : e.g. 'contact@mysite.com'
# MAGICAUTH_LOGGED_IN_REDIRECT_URL_NAME : e.g. 'home'
#
# Note : for the default templates to be picked up, you need to enable
# the app_directories template Loader, by setting TEMPLATES app_dirs to True.
# https://docs.djangoproject.com/en/2.2/ref/templates/api/#django.template.loaders.app_directories.Loader

# Note 2 : magicauth assumes your logout url name is 'logout'.
# If it's not, add MAGICAUTH_LOGOUT_URL_NAME in your settings.

#################
# Email settings
#################
EMAIL_SUBJECT = getattr(django_settings, "MAGICAUTH_EMAIL_SUBJECT", "Lien de connexion")
EMAIL_HTML_TEMPLATE = getattr(
    django_settings, "MAGICAUTH_EMAIL_HTML_TEMPLATE", "magicauth/email.html"
)
EMAIL_TEXT_TEMPLATE = getattr(
    django_settings, "MAGICAUTH_EMAIL_TEXT_TEMPLATE", "magicauth/email.txt"
)
FROM_EMAIL = getattr(django_settings, "MAGICAUTH_FROM_EMAIL")

###########################
# View templates and urls
###########################
# Login view :
# the view in which your user enters their email to start the login process.
LOGIN_URL = getattr(django_settings, "MAGICAUTH_LOGIN_URL", "login/")
# Template for the login view if you want to customise it. It has to contain a form, with a field
# with type="email" and name="email".
LOGIN_VIEW_TEMPLATE = getattr(
    django_settings, "MAGICAUTH_LOGIN_VIEW_TEMPLATE", "magicauth/login.html"
)
# Name of the field in your User model that contains the email
EMAIL_FIELD = getattr(django_settings, "MAGICAUTH_EMAIL_FIELD", "username")

# Email sent view :
# shown when the user has entered their email successfully and the email has been sent.
EMAIL_SENT_VIEW_TEMPLATE = getattr(
    django_settings, "MAGICAUTH_EMAIL_SENT_VIEW_TEMPLATE", "magicauth/email_sent.html"
)
EMAIL_SENT_URL = getattr(django_settings, "MAGICAUTH_EMAIL_SENT_URL", "email-envoyé/")

# Wait view :
# The emailed links point to this url. It shows a wait message, makes the user wait for
# WAIT_SECONDS, and then redirects them to the VALIDATE_TOKEN_URL (see below).
# Why do we have this view? Because some mail clients visit links their find in email, to check them for spam,
# phishing, etc. So that when the user clicks the link in the email, the antispam bot has already "clicked" it first.
# Adding this intermediary view avoids having the antispam bot invalidate the token and block the user login : the bot
# visits the view but does not wait long enough, so the login is not triggered.
WAIT_VIEW_TEMPLATE = getattr(
    django_settings, "MAGICAUTH_WAIT_VIEW_TEMPLATE", "magicauth/wait.html"
)
# The view will look for the token in the "key" variable.
WAIT_URL = getattr(django_settings, "MAGICAUTH_WAIT_URL", "chargement/code/<str:key>/")

# Validate token view :
# validates the token in the url, does the login, and redirects to LOGGED_IN_REDIRECT_URL_NAME.
# This view has no template, the user never sees it.
# The view will look for the token in the "key" variable.
VALIDATE_TOKEN_URL = getattr(
    django_settings, "MAGICAUTH_VALIDATE_TOKEN_URL", "code/<str:key>/"
)

# Logged in redirect view :
# view on which the user lands once logged in. This is a view in your site, probably something
# like "home".
LOGGED_IN_REDIRECT_URL_NAME = getattr(
    django_settings, "MAGICAUTH_LOGGED_IN_REDIRECT_URL_NAME"
)

# Logout view : the url for logout in your site.
LOGOUT_URL_NAME = getattr(django_settings, "MAGICAUTH_LOGOUT_URL_NAME", "logout")

#################
# Other settings
#################
# How long a token stays valid.
# When using an expired token, user will be prompted to get a new one.
TOKEN_DURATION_SECONDS = getattr(
    django_settings, "MAGICAUTH_TOKEN_DURATION_SECONDS", 5 * 60
)
# Function to call when the email entered in the form is not found in the database.
# The default just raises an error whose message gets displayed on the login page.
EMAIL_UNKNOWN_CALLBACK = getattr(
    django_settings, "MAGICAUTH_EMAIL_UNKNOWN_CALLBACK", "magicauth.utils.raise_error"
)
# If using the default EMAIL_UNKNOWN_CALLBACK,
# this message will be displayed when an unknown email is entered.
EMAIL_UNKNOWN_MESSAGE = getattr(
    django_settings, "MAGICAUTH_EMAIL_UNKNOWN_MESSAGE", "Aucun utilisateur trouvé."
)
# How long the user will wait on the WAIT_URL page before doing the actual login.
WAIT_SECONDS = getattr(
    django_settings, "MAGICAUTH_WAIT_SECONDS", 3
)
# This enables the 2FA OTP field
ENABLE_2FA = getattr(
    django_settings, "MAGICAUTH_ENABLE_2FA", False
)
# Can be 6 or 8 (https://django-otp-official.readthedocs.io/en/stable/overview.html#django_otp.plugins.otp_totp.models.TOTPDevice.digits)
OTP_NUM_DIGITS = getattr(
    django_settings, "MAGICAUTH_OTP_NUM_DIGITS", 6
)
if OTP_NUM_DIGITS not in [6, 8] :
    raise ValueError("OTP_NUM_DIGITS must be either 6 or 8 character long --> https://django-otp-official.readthedocs.io/en/stable/overview.html#django_otp.plugins.otp_totp.models.TOTPDevice.digits")




from pytest import mark

from django.shortcuts import reverse
from magicauth import settings
from tests import factories

'''
Step 1 of login process : see doc in magicauth/views.py for step details
'''

pytestmark = mark.django_db


def test_unauthenticated_user_accesses_login_page(client):
    url = reverse("magicauth-login")
    response = client.get(url)
    assert response.status_code == 200


def test_authenticated_user_is_redirected_to_default_redirect_page(client):
    user = factories.UserFactory()
    client.force_login(user)
    url = reverse("magicauth-login")
    response = client.get(url)
    assert response.status_code == 302
    assert response.url == "/landing/"


def test_authenticated_user_is_redirected_to_next_url(client):
    user = factories.UserFactory()
    client.force_login(user)
    url = reverse("magicauth-login")
    response = client.get(url, {"next": "/test_dashboard/"})
    assert response.status_code == 302
    assert response.url == "/test_dashboard/"


def test_login_page_raises_404_if_unsafe_next_url(client):
    url = reverse("magicauth-login")
    response = client.get(url, {"next": "http://www.myfishingsite.com/"})
    assert response.status_code == 404


def test_login_page_raises_404_if_unsafe_next_url_with_authenticated_user(client):
    user = factories.UserFactory()
    client.force_login(user)
    url = reverse("magicauth-login")
    response = client.get(url, {"next": "http://www.myfishingsite.com/"})
    assert user.is_authenticated
    assert response.status_code == 404


def test_template_displays_totp_field_when_2FA_enabled(client):
    settings.ENABLE_2FA = True
    response = client.get(reverse("magicauth-login"))
    assert response.status_code == 200
    assert "Entrez le code" in str(response.content)

def test_template_does_not_displays_totp_field_when_2FA_disabled(client):
    settings.ENABLE_2FA = False
    response = client.get(reverse("magicauth-login"))
    assert response.status_code == 200
    assert "Entrez le code" not in str(response.content)



from django.conf import settings
from django.db import models
from django.utils.translation import ugettext_lazy as _

from .utils import generate_token


class MagicToken(models.Model):
    key = models.CharField(
        verbose_name=_("Key"), primary_key=True, default=generate_token, max_length=255
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, related_name="magic_token", on_delete=models.CASCADE
    )
    created = models.DateTimeField(auto_now_add=True)

    class Meta:
        verbose_name = "Magic Token"
        verbose_name_plural = _("Magic Tokens")
        ordering = ("-created",)

    def __str__(self):
        return self.key



from datetime import timedelta
import logging

from django.contrib import messages
from django.contrib.auth import login
from django.shortcuts import redirect
from django.urls import reverse_lazy, reverse
from django.utils import timezone
from django.views.generic import View, FormView, TemplateView
from django.contrib.auth import get_user_model
from magicauth import settings as magicauth_settings
from magicauth.forms import EmailForm

from magicauth.forms import EmailForm, OTPForm
from magicauth.models import MagicToken
from magicauth.next_url import NextUrlMixin
from magicauth.send_token import SendTokenMixin


logger = logging.getLogger()


class LoginView(NextUrlMixin, SendTokenMixin, FormView):
    """
    Step 1 of login process : GET the LoginView.
    Step 2 of login process : POST your email to the LoginView.
    """

    form_class = EmailForm
    success_url = reverse_lazy("magicauth-email-sent")
    template_name = magicauth_settings.LOGIN_VIEW_TEMPLATE

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            next_url = self.get_next_url(self.request)
            return redirect(next_url)
        return super(LoginView, self).get(request, *args, **kwargs)

    def get_context_data(self, **kwargs):
        context = super(LoginView, self).get_context_data(**kwargs)
        context[
            "LOGGED_IN_REDIRECT_URL_NAME"
        ] = magicauth_settings.LOGGED_IN_REDIRECT_URL_NAME
        context["LOGOUT_URL_NAME"] = magicauth_settings.LOGOUT_URL_NAME
        context["OTP_enabled"] = magicauth_settings.ENABLE_2FA
        context["OTP_form"] = OTPForm(self.request.user)
        return context

    def get_success_url(self, **kwargs):
        url = reverse_lazy("magicauth-email-sent")
        # Use encoded next URL before including it in a string
        next_url_quoted = self.get_next_url_encoded(self.request)
        return f"{url}?next={next_url_quoted}"

    def form_valid(self, form, *args, **kwargs):
        user_email = form.cleaned_data["email"]
        context = {"next_url": self.get_next_url(self.request)}

        def is_OTP_valid():
            if not magicauth_settings.ENABLE_2FA:
                return True
            user = get_user_model().objects.get(email=user_email)
            OTP_form = OTPForm(user=user, data=self.request.POST)
            if OTP_form.is_valid():
                return True
            for error in OTP_form.errors["otp_token"] :
                form.add_error("email", error)
                return False

        if is_OTP_valid():
            self.send_token(user_email=user_email, extra_context=context)
            return super().form_valid(form)

        return super().form_invalid(form)


class EmailSentView(NextUrlMixin, TemplateView):
    """
    Step 3 of login process : you get a confirmation page that the email was sent.
    """
    template_name = magicauth_settings.EMAIL_SENT_VIEW_TEMPLATE


class WaitView(NextUrlMixin, TemplateView):
    """
    Step 4 of login process (optional): you visit the link that you got by email that sends you to
    the WaitView.
    If the WaitView is not used, then the link will send you directly to step 5, token validation.

    Th WaitView will wait few seconds, redirect you to the token validation.
    This is for solving an issue with a security feature in some email clients where
    the magic link is verified and and thus the token gets invalidated by the email client.
    """
    template_name = magicauth_settings.WAIT_VIEW_TEMPLATE

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        token_key = kwargs.get("key")
        validate_token_url = reverse('magicauth-validate-token', kwargs={'key': token_key})
        next_url_quoted = self.get_next_url_encoded(self.request)
        context["next_step_url"] = f"{validate_token_url}?next={next_url_quoted}"
        context["WAIT_SECONDS"] = magicauth_settings.WAIT_SECONDS
        return context


class ValidateTokenView(NextUrlMixin, View):
    """
    Step 5 of login process : you visit the ValidateTokenView that validates the token, logs you in,
    and redirects you to the url in the "next" param (or the default view if no next).

    Either you clicked a link to this page in your email, or you got redirected from step 4
    (WaitView).

    If the token is invalid, you are not logged in, and you are redirected to LoginView (step 1) to
    start over.
    """

    @staticmethod
    def get_valid_token(key):
        duration = magicauth_settings.TOKEN_DURATION_SECONDS
        try:
            token = MagicToken.objects.get(key=key)
        except MagicToken.DoesNotExist:
            return None
        except MagicToken.MultipleObjectsReturned:
            return None

        if token.created < timezone.now() - timedelta(seconds=duration):
            token.delete()
            return None
        return token

    def get(self, request, *args, **kwargs):
        if request.user.is_authenticated:
            url = self.get_next_url(request)
            return redirect(url)
        token_key = kwargs.get("key")
        token = self.get_valid_token(token_key)
        if not token:
            messages.warning(
                self.request,
                "Ce lien de connexion ne fonctionne plus. "
                "Pour en recevoir un nouveau, nous vous invitons à renseigner "
                "votre email ci-dessous puis à cliquer sur valider.",
            )
            return redirect("magicauth-login")
        url = self.get_next_url(request)
        login(self.request, token.user)
        MagicToken.objects.filter(
            user=token.user
        ).delete()  # Remove them all for this user
        return redirect(url)
